use std::collections::HashMap;

use crate::model::{
    role::Role,
    user::{AdminUser, User, UserWithRoles},
};
use futures::future::join_all;
use rocket::{
    form::Form,
    get, post,
    request::FlashMessage,
    response::{Flash, Redirect},
    routes, FromForm, Route, State,
};
use rocket_dyn_templates::{tera::Context, Template};
use sqlx::SqlitePool;

#[get("/user")]
async fn index(
    db: &State<SqlitePool>,
    admin: AdminUser,
    flash: Option<FlashMessage<'_>>,
) -> Template {
    let user_futures: Vec<_> = User::all(db)
        .await
        .into_iter()
        .map(|u| async move { UserWithRoles::from_user(u, db).await })
        .collect();

    let users: Vec<UserWithRoles> = join_all(user_futures).await;

    let roles = Role::all(db).await;

    let mut context = Context::new();
    if let Some(msg) = flash {
        context.insert("flash", &msg.into_inner());
    }
    context.insert("users", &users);
    context.insert("roles", &roles);
    context.insert(
        "loggedin_user",
        &UserWithRoles::from_user(admin.user, db).await,
    );

    Template::render("admin/user/index", context.into_json())
}

#[get("/user/<user>/reset-pw")]
async fn resetpw(db: &State<SqlitePool>, _admin: AdminUser, user: i32) -> Flash<Redirect> {
    let user = User::find_by_id(db, user).await;
    match user {
        Some(user) => {
            user.reset_pw(db).await;
            Flash::success(
                Redirect::to("/admin/user"),
                format!("Passwort von {} zurückgesetzt", user.name),
            )
        }
        None => Flash::error(Redirect::to("/admin/user"), "User does not exist"),
    }
}

#[get("/user/<user>/delete")]
async fn delete(db: &State<SqlitePool>, _admin: AdminUser, user: i32) -> Flash<Redirect> {
    let user = User::find_by_id(db, user).await;
    match user {
        Some(user) => {
            user.delete(db).await;
            Flash::success(
                Redirect::to("/admin/user"),
                format!("Benutzer {} gelöscht", user.name),
            )
        }
        None => Flash::error(Redirect::to("/admin/user"), "User does not exist"),
    }
}

#[derive(FromForm, Debug)]
pub struct UserEditForm {
    pub(crate) id: i32,
    pub(crate) dob: Option<String>,
    pub(crate) weight: Option<String>,
    pub(crate) sex: Option<String>,
    pub(crate) roles: HashMap<String, String>,
    pub(crate) member_since_date: Option<String>,
    pub(crate) birthdate: Option<String>,
    pub(crate) mail: Option<String>,
    pub(crate) nickname: Option<String>,
    pub(crate) notes: Option<String>,
    pub(crate) phone: Option<String>,
    pub(crate) address: Option<String>,
}

#[post("/user", data = "<data>")]
async fn update(
    db: &State<SqlitePool>,
    data: Form<UserEditForm>,
    _admin: AdminUser,
) -> Flash<Redirect> {
    let user = User::find_by_id(db, data.id).await;
    let Some(user) = user else {
        return Flash::error(
            Redirect::to("/admin/user"),
            format!("User with ID {} does not exist!", data.id),
        );
    };

    user.update(db, data.into_inner()).await;

    Flash::success(Redirect::to("/admin/user"), "Successfully updated user")
}

#[derive(FromForm)]
struct UserAddForm<'r> {
    name: &'r str,
}

#[post("/user/new", data = "<data>")]
async fn create(
    db: &State<SqlitePool>,
    data: Form<UserAddForm<'_>>,
    _admin: AdminUser,
) -> Flash<Redirect> {
    if User::create(db, data.name).await {
        Flash::success(Redirect::to("/admin/user"), "Successfully created user")
    } else {
        Flash::error(
            Redirect::to("/admin/user"),
            format!("User {} already exists", data.name),
        )
    }
}

pub fn routes() -> Vec<Route> {
    routes![index, resetpw, update, create, delete]
}
